/*
 * bzlib_wrapper.c -- bzlib wrapper
 * (C)Copyright 1999 by Hiroshi Takekawa
 * This file is part of Enfle.
 *
 * Last Modified: Mon May  1 01:56:30 2000.
 * $Id: bzlib_wrapper.c,v 1.2 2000/05/08 09:53:21 sian Exp $
 *
 * Enfle is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * Enfle is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA
 */

#include <stdio.h>
#include <bzlib.h>
#include "enfle.h"
#include "bz2.h"
#include "bzlib_wrapper.h"

static int
bzseek_set(Archive *ar, long pos)
{
  BZ2_info *info = ar->info;
  unsigned char buf[BZLIB_WRAPPER_BUFSIZE];
  int bytes_to_read;

  if (info->pos < pos) {
    bytes_to_read = pos - info->pos;
  } else {
    BZCLOSE(info->bzfile);
    if ((info->bzfile = BZOPEN(ar->filename, "rb")) == NULL)
      return -1;
    info->pos = 0;
    bytes_to_read = pos;
  }
  while (bytes_to_read > BZLIB_WRAPPER_BUFSIZE) {
    if (BZREAD(info->bzfile, buf, BZLIB_WRAPPER_BUFSIZE) < 0)
      return -1;
    info->pos += BZLIB_WRAPPER_BUFSIZE;
      bytes_to_read -= BZLIB_WRAPPER_BUFSIZE;
  }
  if (BZREAD(info->bzfile, buf, bytes_to_read) < 0)
    return -1;
  info->pos += bytes_to_read;
  return info->pos;
}

int
bzseek(Archive *ar, long pos, int whence)
{
  BZ2_info *info = ar->info;

  switch (whence) {
  case SEEK_SET:
    return bzseek_set(ar, pos);
  case SEEK_CUR:
    return bzseek_set(ar, info->pos + pos);
  case SEEK_END:
    /* SEEK_END cannot be implemented... */
    return 0;
  default:
    fprintf(stderr, "bzseek: Unknown whence\n");
    return -1;
  }
}
