/*
 * ald.c -- ald file handler
 * (C)Copyright 1998, 99 by Hiroshi Takekawa
 * This file is part of Enfle.
 *
 * Last Modified: Sun Sep 26 16:56:14 1999.
 * $Id: ald.c,v 1.10 1999/09/26 08:55:36 sian Exp $
 *
 * by courtesy of xsystem3.5
 *
 * Enfle is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * Enfle is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA
 */

#include "cmrender.h"
#include "quantize.h"
#include "pms.h"
#include "vsp.h"
#include "ald.h"
#include "plugin.h"

int
#ifdef PIC
get_plugininfo(PluginInfo *p)
#else
loader_ald_get_plugininfo(PluginInfo *p)
#endif
{
  p->version = 1;
  p->type = _Loader;
  p->pluginname = "ALD(PMS,VSP) Format Loader Plugin version 0.1.2";
  p->pluginshortname = FORMAT_NAME;
  p->author = "Hiroshi Takekawa";
  p->dlhandle = NULL; /* set by plugin_load */
  p->functions.loader.load_image = ald_load_image;

  return 1;
}

int
ald_load_image(Info *info, Image *p)
{
  PMS_info *pmsinfo;
  VSP_info *vspinfo;
  unsigned char *data;

#ifdef DEBUG
  fprintf(stderr, "ald_load_image(): allocate memory %d bytes\n", info->ar->size);
#endif
  if ((data = malloc(info->ar->size)) == NULL)
    return PROC_ERROR;
  archive_read(info->ar, data, info->ar->size);
  if ((pmsinfo = PMS_load(data)) != NULL) {
    if (pmsinfo->depth == 8)
      memcpy(p->colormap, pmsinfo->palette, 3 * 256);

    p->width = pmsinfo->width;
    p->height = pmsinfo->height;

    p->type = (pmsinfo->depth == 16 || pmsinfo->depth == 15) ?
      _RGB16 : _INDEX;
    p->ncolors = 1 << pmsinfo->depth;
    info->npics = info->ar->nfiles;

    p->image = pmsinfo->image;
    p->bytes_per_line = p->width * pmsinfo->depth / 8;
    p->image_size = p->bytes_per_line * p->height;
    if (pmsinfo->comment != NULL)
      info->comment = strdup(pmsinfo->comment);
    free(pmsinfo);
  } else if ((vspinfo = VSP_load(data)) != NULL) {
    memcpy(p->colormap, vspinfo->palette, 3 * 16);

    p->width = vspinfo->width;
    p->height = vspinfo->height;

    p->type = _INDEX;
    p->ncolors = 16;
    info->npics = info->ar->nfiles;

    p->image = vspinfo->image;
    p->bytes_per_line = p->width;
    p->image_size = p->bytes_per_line * p->height;
    free(vspinfo);
  } else {
#ifdef DEBUG
  fprintf(stderr, "ald_load_image(): failed\n");
#endif
    free(data);
    return PROC_NOT;
  }
#ifdef DEBUG
  fprintf(stderr, "done.\n");
#endif
  free(data);

  info->format = FORMAT_NAME;
  info->handler.destroy  = NULL;
  info->handler.alarmset = NULL;
  info->handler.alarm    = NULL;
  info->handler.alarmoff = NULL;

  return PROC_OK;
}
