/*
 * pms.c -- pms file expander
 * (C)Copyright 1998 by Hiroshi Takekawa
 * This file if part of Enfle.
 *
 * Last Modified: Mon Sep 13 06:34:11 1999.
 * $Id: pms.c,v 1.3 1999/09/15 17:29:53 sian Exp $
 *
 * Enfle is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * Enfle is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA
 */

#ifdef HAVE_CONFIG_H
#  include "config.h"
#endif
#include "pms.h"
#include "utils.h"

#include <stdio.h>
#include <stdlib.h>
#if STDC_HEADERS
#  include <string.h>
#else
#  ifndef HAVE_MEMCPY
#    define memcpy(d, s, n) bcopy ((s), (d), (n))
#    define memmove(d, s, n) bcopy ((s), (d), (n))
#  endif
#endif

/* #define DEBUG */
/* #define ALPHA */

PMS_info *
pms_loadheader(unsigned char *d)
{
  PMS_info *p = malloc(sizeof(PMS_info));

  if (p == NULL)
    return NULL;

#ifdef DEBUG
  fprintf(stderr, "checking signature...");
#endif
  if (d[0] != 'P' || d[1] != 'M') {
    free(p);
    return NULL;
  }
#ifdef DEBUG
  fprintf(stderr, "ok\n");
#endif

  p->version = get_little_word(d + 2);
  p->headersize = get_little_word(d + 4);
  p->depth = d[6];
  p->depths = d[7];
  p->sf = d[8];
  p->bf = get_little_word(d + 10);
  p->x = get_little_dword(d + 16);
  p->y = get_little_dword(d + 20);
  p->width = get_little_dword(d + 24);
  p->height = get_little_dword(d + 28);
  p->d_offset = get_little_dword(d + 32);
  p->p_offset = get_little_dword(d + 36);
  p->c_offset = get_little_dword(d + 40);
  p->comment = d[0x30] != '\0' ? d + 48 : NULL;

  return p;
}

unsigned char *
pms_decode_256(unsigned char *d, int w, int h)
{
  unsigned char *p, *ptr, c, c2;
  int x, y, l, i;

  if ((p = ptr = malloc(w * h)) == NULL)
    return NULL;

  for (y = 0; y < h; y++) {
    for (x = 0; x < w;) {
      c = *d++;
      if (c < 0xf8) {
	*p++ = c;
	x++;
      } else if (c == 0xff) {
	l = (*d++) + 3;
	memcpy(p, p - w, l);
	x += l;
	p += l;
      } else if (c == 0xfe) {
	l = (*d++) + 3;
	memcpy(p, p - w * 2, l);
	x += l;
	p += l;
      } else if (c == 0xfd) {
	l = (*d++) + 4;
	memset(p, *d++, l);
	x += l;
	p += l;
      } else if (c == 0xfc) {
	l = ((*d++) + 3) * 2;
	c = *d++;
	c2 = *d++;
	for (i = 0; i < l; i += 2) {
	  *p++ = c;
	  *p++ = c2;
	}
	x += l;
      } else {
	*p++ = *d++;
	x++;
      }
    }
  }

  return ptr;
}

unsigned char *
pms_decode_64k(unsigned char *d, int w, int h, unsigned char *alpha)
{
  unsigned char *p, *ptr;
  int c, c2, c3, c4;
  int x, y, l, i, pc0, pc1;

  if ((p = ptr = malloc(w * h * 2)) == NULL)
    return NULL;

  for (y = 0; y < h; y++) {
    for (x = 0; x < w;) {
      c = *d++;
      if (c < 0xf8) {
	*p++ = c;
	*p++ = *d++;
	x++;
      } else if (c == 0xff) {
	l = (*d++) + 2;
	memcpy(p, p - w * 2, l * 2);
	x += l;
	p += l * 2;
      } else if (c == 0xfe) {
	l = (*d++) + 2;
	memcpy(p, p - w * 4, l * 2);
	x += l;
	p += l * 2;
      } else if (c == 0xfd) {
	l = (*d++) + 3;
	c = *d++;
	c2 = *d++;
	for (i = 0; i < l; i++) {
	  *p++ = c;
	  *p++ = c2;
	}
	x += l;
      } else if (c == 0xfc) {
	l = ((*d++) + 2) * 2;
	c = *d++;
	c2 = *d++;
	c3 = *d++;
	c4 = *d++;
	for (i = 0; i < l; i += 2) {
	  *p++ = c;
	  *p++ = c2;
	  *p++ = c3;
	  *p++ = c4;
	}
	x += l;
      } else if (c == 0xfb) {
	*p++ = *(p - w * 2 - 2);
	*p++ = *(p - w * 2 - 2);
	x++;
      } else if (c == 0xfa) {
	*p++ = *(p - w * 2 + 2);
	*p++ = *(p - w * 2 + 2);
	x++;
      } else if (c == 0xf9) {
	l = (*d++) + 1;
	c = *d++;
	c2 = *d++;
	pc0 = ((c & 0xe0) << 8) + ((c & 0x18) << 6) + ((c & 0x07) << 2);
	pc1 = ((c2 & 0xc0) << 5) + ((c2 & 0x3c) << 3) + (c2 & 0x03);
	*p++ = (pc0 | pc1) & 0xff;
	*p++ = (pc0 | pc1) >> 8;
	for (i = 1; i < l; i++) {
	  c2 = *d++;
	  pc1 = ((c2 & 0xc0) << 5) + ((c2 & 0x3c) << 3) + (c2 & 0x03);
	  *p++ = (pc0 | pc1) & 0xff;
	  *p++ = (pc0 | pc1) >> 8;
	}
	x += l;
      } else {
	*p++ = *d++;
	*p++ = *d++;
	x++;
      }
    }
  }

#ifdef ALPHA
  if (alpha != NULL) {
    unsigned char *a, *ap;
    unsigned short int *p;
    a = ap = pms_decode_256(alpha, w, h);
    p = (unsigned short int *)ptr;
    /* alpha compound */
    for (y = 0; y < h; y++)
      for (x = 0; x < w; x++) {
	unsigned int pixel;
	if (*a < 255) {
	  pixel = *p;
	  *p = PIX16((PIX16R(pixel) * (*a)) >> 8,
		     (PIX16G(pixel) * (*a)) >> 8,
		     (PIX16B(pixel) * (*a)) >> 8);
	}
	p++;
	a++;
      }
    free(ap);
  }
#endif

  return ptr;
}

PMS_info *
PMS_load(unsigned char *data)
{
  PMS_info *pmsinfo;

  if (data == NULL)
    return NULL;

  if ((pmsinfo = pms_loadheader(data)) == NULL)
    return NULL;
#ifdef DEBUG
  fprintf(stderr, "width = %d, height = %d\n", pmsinfo->width, pmsinfo->height);
#endif

  if (pmsinfo->depth == 8) {
    memcpy(pmsinfo->palette, data + pmsinfo->p_offset, 3 * 256);
    pmsinfo->image = pms_decode_256(data + pmsinfo->d_offset,
				    pmsinfo->width, pmsinfo->height);
  } else if (pmsinfo->depth == 16)
    pmsinfo->image = pms_decode_64k(data + pmsinfo->d_offset,
				    pmsinfo->width, pmsinfo->height,
				    pmsinfo->p_offset ? data + pmsinfo->p_offset : NULL);
  else
    fprintf(stderr, "Not supported\n");

  return pmsinfo;
}
