/*
 * lf2.c -- lf2 file handler
 * (C)Copyright 1999 by Hiroshi Takekawa
 * This file is part of Enfle.
 *
 * Last Modified: Tue Sep 21 07:51:30 1999.
 * $Id: lf2.c,v 1.8 1999/09/20 23:48:27 sian Exp $
 *
 * by courtesy of TF
 *
 * Enfle is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * Enfle is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA
 */

#include "lf2.h"
#include "utils.h"
#include "plugin.h"

static int lf2_read_image(Info *info, Image *p)
{
  Archive *ar = info->ar;
  unsigned char ring[RING_BUFFER_SIZE], flag = 0, data, upper, lower;
  int i, j, c, index, pos, len, m;

  memset(ring, 0, sizeof(ring));

  if ((p->image = malloc(p->image_size)) == NULL) {
    fprintf(stderr, "lf2_read_image: No enough memory for image\n");
    return 0;
  }

  for (i = 0, c = 0, m = 0xfee; i < p->image_size;) {
    /* flag bits */
    if (--c < 0) {
      flag = archive_getc(ar) ^ 0xff;
      c = 7;
    }

    if (flag & 0x80) {
      /* data */
      ring[m] = data = archive_getc(ar) ^ 0xff;
      m = (++m) % RING_BUFFER_SIZE;
      index = (i % p->width) + p->width * (p->height - i / p->width - 1);
      p->image[index] = data;
      i++;
    } else {
      /* copy */
      upper = archive_getc(ar) ^ 0xff;
      lower = archive_getc(ar) ^ 0xff;
      len = (upper & 0xf) + 3;
      pos = (upper >> 4) + (lower << 4);
      for (j = 0; j < len; j++) {
	index = (i % p->width) + p->width * (p->height - i / p->width - 1);
	ring[m] = p->image[index] = ring[pos];
	m = (++m) % RING_BUFFER_SIZE;
	pos = (++pos) % RING_BUFFER_SIZE;
	i++;
      }
    }
    flag <<= 1;
  }

  return 1;
}

int lf2_decode_image(Info *info, Image *p)
{
  int i;
  char buf[8];
  unsigned char dummy, tmp;
  Archive *ar = info->ar;

  if (archive_read(ar, buf, 8) != 8)
    return 0;

  if (memcmp("LEAF256\0", buf, 8))
    return 0;

  info->format = FORMAT_NAME;

#ifdef DEBUG
  fprintf(stderr, "lf2_decode_image: lf2 identified\n");
#endif

  /* read size information */
  p->left = read_little_word(ar);
  p->top = read_little_word(ar);
  p->width = read_little_word(ar);
  p->height = read_little_word(ar);

  /* set other variables */
  p->type = _INDEX;
  p->bytes_per_line = p->width;
  p->image_size = p->bytes_per_line * p->height;

  /* skip unknown(or padding) */
  dummy = archive_getc(ar);
  dummy = archive_getc(ar);

  /* read transparent */
  if ((tmp = archive_getc(ar)) < 0xff) {
    p->transparent_disposal = info->transparent_disposal;
    p->transparent.index = tmp;
  } else
    p->transparent_disposal = _DONOTHING;

  /* skip padding */
  dummy = archive_getc(ar);
  dummy = archive_getc(ar);
  dummy = archive_getc(ar);

  /* read # of colors */
  p->ncolors = archive_getc(ar);

  /* skip padding */
  dummy = archive_getc(ar);

  for (i = 0; i < p->ncolors; i++) {
    p->colormap[i][2] = archive_getc(ar);
    p->colormap[i][1] = archive_getc(ar);
    p->colormap[i][0] = archive_getc(ar);
  }

  return lf2_read_image(info, p);
}
