/*
 * png.c -- png driver for enfle, using libpng
 * (C)Copyright 1998, 99 by Hiroshi Takekawa
 * This file is part of Enfle.
 *
 * Last Modified: Sat Sep  9 08:02:40 2000.
 * $Id: png.c,v 1.14 2000/09/08 23:12:11 sian Exp $
 *
 * Note: png implementation is far from complete.
 *
 * Enfle is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * Enfle is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA
 */

#include <png.h>

#include "enfle.h"
#include "_png.h"
#include "quantize.h"
#include "cmrender.h"
#include "plugin.h"

#ifndef TRUE
#  define TRUE 1
#  define FALSE 0
#endif

#define DISPLAY_GAMMA 2.20

int
#ifdef PIC
get_plugininfo(PluginInfo *p)
#else
loader_png_get_plugininfo(PluginInfo *p)
#endif
{
  p->version = 1;
  p->type = _Loader;
  p->pluginname = "PNG Format Loader Plugin version 0.2";
  p->pluginshortname = FORMAT_NAME;
  p->author = "Hiroshi Takekawa";
  p->dlhandle = NULL; /* set by plugin_load */
  p->functions.loader.load_image = png_load_image;

  return 1;
}

static int
png_process_alpha_rgb(Image *p)
{
  int x, y, bit;
  unsigned char *ptr = p->image, *d, *dest, *md = NULL, alpha, b;
  const int bit_mask[8] = {128, 64, 32, 16, 8, 4, 2, 1};

  if ((d = dest = malloc(p->width * 3 * p->height)) == NULL)
    return 0;

#ifdef SHAPE
  p->mask_size = (int)((p->width + 7) >> 3) * p->height;
  if ((p->mask = md = malloc(p->mask_size)) == NULL) {
    free(d);
    return 0;
  }
#endif

  for (y = 0; y < p->height; y++) {
    bit = 0;
    b = 0;
    for (x = 0; x < p->width; x++) {
      alpha = ptr[3];
      if (alpha == 0)
	b |= bit_mask[bit];
      bit++;
      if (bit == 8) {
#ifdef SHAPE
	*md++ = b;
#endif
	bit = 0;
	b = 0;
      }
      d[0] = (ptr[0] - p->background.red)   * (alpha / 255) + p->background.red;
      d[1] = (ptr[1] - p->background.green) * (alpha / 255) + p->background.green;
      d[2] = (ptr[2] - p->background.blue)  * (alpha / 255) + p->background.blue;
      d += 3;
      ptr += 4;
    }
#ifdef SHAPE
    if (bit > 0)
      *md++ = b;
#endif
  }

  p->transparent.red = p->background.red;
  p->transparent.green = p->background.green;
  p->transparent.blue = p->background.blue;

  free(p->image);
  p->image = dest;
  p->bytes_per_line = p->width * 3;
  p->image_size = p->bytes_per_line * p->height;

  return 1;
}

static int
png_process_alpha_gray(Image *p)
{
  int x, y, bit;
  unsigned char *ptr = p->image, *d, *dest, *md = NULL, alpha, b;
  const int bit_mask[8] = {128, 64, 32, 16, 8, 4, 2, 1};

  if ((d = dest = malloc(p->width * p->height)) == NULL)
    return 0;

#ifdef SHAPE
  p->mask_size = (int)((p->width + 7) / 8) * p->height;
  if ((p->mask = md = malloc(p->mask_size)) == NULL) {
    free(d);
    return 0;
  }
#endif

  for (y = 0; y < p->height; y++) {
    bit = 0;
    b = 0;
    for (x = 0; x < p->width; x++) {
      alpha = ptr[1];
      if (alpha == 0)
	b |= bit_mask[bit];
      bit++;
      if (bit == 8) {
#ifdef SHAPE
	*md++ = b;
#endif
	bit = 0;
	b = 0;
      }
      d[0] = (ptr[0] - p->background.index)   * (alpha / 255) + p->background.index;
      d++;
      ptr += 2;
    }
#ifdef SHAPE
    if (bit > 0)
      *md++ = b;
#endif
  }

  p->transparent.red = p->background.red;
  p->transparent.green = p->background.green;
  p->transparent.blue = p->background.blue;

  free(p->image);
  p->image = dest;
  p->bytes_per_line = p->width * 3;
  p->image_size = p->bytes_per_line * p->height;

  return 1;
}

static void
png_error_handler(png_structp png_ptr, png_const_charp error_msg)
{
  int *try_when_error = (int *)png_get_error_ptr(png_ptr);

  if (!strcmp(error_msg, "incorrect data check"))
    *try_when_error = 1;
  else
    fprintf(stderr, "enfle: png error: %s\n", error_msg);

  longjmp(png_jmpbuf(png_ptr), 1);
}

static void
png_warning_handler(png_structp png_ptr, png_const_charp warning_msg)
{
  /* fprintf(stderr, "enfle: png warning: %s\n", warning_msg); */
}

static void
png_read_data(png_structp png_ptr, png_bytep data, png_uint_32 len)
{
  Archive *ar = png_get_io_ptr(png_ptr);

  archive_read(ar, data, len);
}

#define PNG_BYTES_TO_CHECK 4
static int
png_decode_image(Info *info, Image *p)
{
  Archive *ar = info->ar;
  char buf[PNG_BYTES_TO_CHECK];
  png_structp png_ptr;
  png_infop info_ptr;
  png_uint_32 width, height;
  png_color_16 my_background, *image_background;
  png_color_16p trans_values;
  png_bytep *image_array, trans;
  png_textp text_ptr;
  int bit_depth, color_type, interlace_type;
  int i, num_trans, num_text, text_len;
  int try_when_error;

#ifdef DEBUG
  fprintf(stderr, "png_decode_image() called\n");
#endif

  /* Read in the signature bytes */
  if (archive_read(ar, buf, PNG_BYTES_TO_CHECK) != PNG_BYTES_TO_CHECK)
    return PROC_NOT;

  /* Compare the first PNG_BYTES_TO_CHECK bytes of the signature. */
  if (png_sig_cmp(buf, (png_size_t)0, PNG_BYTES_TO_CHECK) != 0)
    return PROC_NOT;

  try_when_error = 0;
  /* Create and initialize the png_struct */
  if ((png_ptr = png_create_read_struct(PNG_LIBPNG_VER_STRING, (void *)&try_when_error,
					png_error_handler, png_warning_handler)) == NULL)
    return PROC_ERROR;

  /* Allocate/initialize the memory for image information. */
  if ((info_ptr = png_create_info_struct(png_ptr)) == NULL) {
    png_destroy_read_struct(&png_ptr, (png_infopp)NULL, (png_infopp)NULL);
    return PROC_ERROR;
  }

  if (setjmp(png_jmpbuf(png_ptr))) {
    png_destroy_read_struct(&png_ptr, &info_ptr, (png_infopp)NULL);
    return try_when_error ? PROC_OK : PROC_ERROR;
  }

  png_set_read_fn(png_ptr, (voidp)ar, (png_rw_ptr)png_read_data);
  png_set_sig_bytes(png_ptr, PNG_BYTES_TO_CHECK);
  png_read_info(png_ptr, info_ptr);
  png_get_IHDR(png_ptr, info_ptr, &width, &height, &bit_depth, &color_type,
	       &interlace_type, NULL, NULL);

#ifdef DEBUG
  switch (color_type) {
  case PNG_COLOR_TYPE_PALETTE:
    fprintf(stderr, "PALETTE\n");
    break;
  case PNG_COLOR_TYPE_RGB:
    fprintf(stderr, "RGB\n");
    break;
  case PNG_COLOR_TYPE_RGB_ALPHA:
    fprintf(stderr, "RGB_ALPHA\n");
    break;
  case PNG_COLOR_TYPE_GRAY:
    fprintf(stderr, "GRAY\n");
    break;
  case PNG_COLOR_TYPE_GRAY_ALPHA:
    fprintf(stderr, "GRAY_ALPHA\n");
    break;
  default:
    fprintf(stderr, "UNKNOWN\n");
    break;
  }
#endif

  info->format = FORMAT_NAME;
  p->width = width;
  p->height = height;

  /* read comment */
  if ((num_text = png_get_text(png_ptr, info_ptr, &text_ptr, 0)) > 0) {
    text_len = 0;
    for (i = 0; i < num_text; i++)
      text_len += strlen(text_ptr[i].key) + 2 + strlen(text_ptr[i].text) + 1;
    if ((info->comment = malloc(text_len + 1)) == NULL) {
      png_destroy_read_struct(&png_ptr, &info_ptr, (png_infopp)NULL);
      return PROC_ERROR;
    }
    memset(info->comment, 0, text_len + 1);
    for (i = 0; i < num_text; i++) {
      strcat(info->comment, text_ptr[i].key);
      strcat(info->comment, ": ");
      strcat(info->comment, text_ptr[i].text);
      strcat(info->comment, "\n");
    }
  }

  if (bit_depth == 16)
    png_set_strip_16(png_ptr);
  else if (bit_depth < 8)
    png_set_packing(png_ptr);

  if (color_type == PNG_COLOR_TYPE_GRAY && bit_depth < 8)
    png_set_expand(png_ptr);

  png_read_update_info(png_ptr, info_ptr);

  /* Set the transparent color */
  if (png_get_tRNS(png_ptr, info_ptr, &trans, &num_trans, &trans_values)) {
    int i;

    p->transparent_disposal = info->transparent_disposal;
    switch (color_type) {
    case PNG_COLOR_TYPE_PALETTE:
      for (i = 0; i < num_trans; i++)
	if (trans[i] == 0) {
	  p->transparent.index = i;
	  break;
	}
      break;
    case PNG_COLOR_TYPE_RGB:
    case PNG_COLOR_TYPE_RGB_ALPHA:
      p->transparent.red = info_ptr->trans_values.red >> 8;
      p->transparent.green = info_ptr->trans_values.green >> 8;
      p->transparent.blue = info_ptr->trans_values.blue >> 8;
      break;
    case PNG_COLOR_TYPE_GRAY:
    case PNG_COLOR_TYPE_GRAY_ALPHA:
      p->transparent.index = info_ptr->trans_values.gray;
      break;
    default:
      break;
    }
  }

  /* Set the background color */
  if (color_type != PNG_COLOR_TYPE_RGB_ALPHA && color_type != PNG_COLOR_TYPE_GRAY_ALPHA) {
    if (png_get_bKGD(png_ptr, info_ptr, &image_background)) {
      p->background.red = image_background->red;
      p->background.green = image_background->green;
      p->background.blue = image_background->blue;
      p->background.index = image_background->gray;
      png_set_background(png_ptr, image_background, PNG_BACKGROUND_GAMMA_FILE, 1, 1.0);
    } else {
      if (num_trans) {
	p->background.red   = my_background.red   = p->transparent.red;
	p->background.green = my_background.green = p->transparent.green;
	p->background.blue  = my_background.blue  = p->transparent.blue;
	p->background.index = my_background.index = p->transparent.index;
      } else {
	p->background.red = p->background.green = p->background.blue =
	  p->background.index =
	  my_background.red = my_background.green = my_background.blue =
	  my_background.gray = 0;
      }
      png_set_background(png_ptr, &my_background, PNG_BACKGROUND_GAMMA_SCREEN, 0, DISPLAY_GAMMA);
    }
  } else {
    if (!png_get_bKGD(png_ptr, info_ptr, &image_background))
      p->background.red = p->background.green =	p->background.blue = p->background.index = 0;
  }

  /* prepare image data, store palette if exists */
  switch (color_type) {
  case PNG_COLOR_TYPE_RGB:
    p->type = _RGB24;
    p->ncolors = 1 << 24;
    p->bytes_per_line = width * 3;
    p->image_size = p->bytes_per_line * height;
    break;
  case PNG_COLOR_TYPE_RGB_ALPHA:
    p->type = _RGB24;
    p->ncolors = 1 << 24;
    p->bytes_per_line = width * 4;
    p->image_size = p->bytes_per_line * height;
    break;
  case PNG_COLOR_TYPE_GRAY:
    p->type = _INDEX;
    p->ncolors = 256;
    p->bytes_per_line = width;
    p->image_size = p->bytes_per_line * height;

    for (i = 0; i < 256; i++) {
      p->colormap[i][0] =
	p->colormap[i][1] =
	p->colormap[i][2] = i;
    }
    break;
  case PNG_COLOR_TYPE_GRAY_ALPHA:
    p->type = _INDEX;
    p->ncolors = 256;
    p->bytes_per_line = width * 2;
    p->image_size = p->bytes_per_line * height;

    for (i = 0; i < 256; i++) {
      p->colormap[i][0] =
	p->colormap[i][1] =
	p->colormap[i][2] = i;
    }
    break;
  default:
    p->type = _INDEX;
    p->ncolors = info_ptr->num_palette;
    p->bytes_per_line = width;
    p->image_size = width * height;

    for (i = 0; i < p->ncolors; i++) {
      p->colormap[i][0] = info_ptr->palette[i].red;
      p->colormap[i][1] = info_ptr->palette[i].green;
      p->colormap[i][2] = info_ptr->palette[i].blue;
    }
  }

  /* allocate memory for returned image */
  if ((p->image = calloc(1, p->image_size)) == NULL) {
    png_destroy_read_struct(&png_ptr, &info_ptr, (png_infopp)NULL);
    return PROC_ERROR;
  }

  /* allocate memory for pointer array */
  if ((image_array = calloc(height, sizeof(png_bytep))) == NULL) {
    png_destroy_read_struct(&png_ptr, &info_ptr, (png_infopp)NULL);
    free(p->image);
    p->image = NULL;
    return PROC_ERROR;
  }

  for (i = 0; i < height; i++)
    image_array[i] = p->image + png_get_rowbytes(png_ptr, info_ptr) * i;

  /* read image */
  png_read_image(png_ptr, image_array);

  /* read rest of file, and get additional chunks in info_ptr */
  png_read_end(png_ptr, info_ptr);

  /* process alpha channel if exists */
  if (color_type == PNG_COLOR_TYPE_RGB_ALPHA) {
    if (!png_process_alpha_rgb(p)) {
      png_destroy_read_struct(&png_ptr, &info_ptr, (png_infopp)NULL);
      free(p->image);
      p->image = NULL;
      return PROC_ERROR;
    } else
      p->transparent_disposal = info->transparent_disposal;
  } else if (color_type == PNG_COLOR_TYPE_GRAY_ALPHA) {
    if (!png_process_alpha_gray(p)) {
      png_destroy_read_struct(&png_ptr, &info_ptr, (png_infopp)NULL);
      free(p->image);
      p->image = NULL;
      return PROC_ERROR;
    } else
      p->transparent_disposal = info->transparent_disposal;
  }

  /* clean up after the read, and free any memory allocated */
  png_destroy_read_struct(&png_ptr, &info_ptr, (png_infopp)NULL);

  /* free temporary memory */
  free(image_array);

  return PROC_OK;
}

int
png_load_image(Info *info, Image *p)
{
  int f;

  if ((f = png_decode_image(info, p)) != PROC_OK)
    return f;

  /* set handler */
  info->handler.destroy  = NULL;
  info->handler.alarmset = NULL;
  info->handler.alarm    = NULL;
  info->handler.alarmoff = NULL;

  info->npics = 1;

  return PROC_OK;
}
