/*
 * ungif.c -- Enfle yet another GIF handler, this uses libungif
 * (C)Copyright 1998, 99 by Hiroshi Takekawa
 * This file is part of Enfle.
 *
 * Last Modified: Tue Oct 10 21:01:41 2000.
 * $Id: ungif.c,v 1.4 2000/10/10 17:03:45 sian Exp $
 *
 * NOTE: This file does not include LZW code.
 * NOTE: This plugin is incomplete, but works.
 *
 *             The Graphics Interchange Format(c) is
 *       the Copyright property of CompuServe Incorporated.
 * GIF(sm) is a Service Mark property of CompuServe Incorporated.
 *
 * Enfle is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * Enfle is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA
 */

#include <gif_lib.h>
#include <signal.h>
#include <sys/time.h>

#include "enfle.h"
#include "ungif.h"
#include "plugin.h"

int
#ifdef PIC
get_plugininfo(PluginInfo *p)
#else
loader_ungif_get_plugininfo(PluginInfo *p)
#endif
{
  p->version = 1;
  p->type = _Loader;
  p->pluginname = "UNGIF Format Loader Plugin version 0.2.1";
  p->pluginshortname = FORMAT_NAME;
  p->author = "Hiroshi Takekawa";
  p->dlhandle = NULL; /* set by plugin_load */
  p->functions.loader.load_image = ungif_load_image;

  return 1;
}

static int
ungif_input_func(GifFileType *GifFile, GifByteType *p, int s)
{
  Archive *ar = (Archive *)GifFile->UserData;

  return archive_read(ar, p, s);
}

static int
ungif_decode_image(Info *info, Image *p)
{
  int i, j, size, sheight, row, col, extcode;
  static int ioffset[] = { 0, 4, 2, 1 };
  static int ijumps[] = { 8, 8, 4, 2 };
  GifRecordType RecordType;
  GifByteType *Extension;
  GifRowType *ScreenBuffer;
  GifFileType *GifFile;
  ColorMapObject *ColorMap;
  Image *q = NULL, *r;
  Transparent transparent_disposal = _DONOTHING;
  int transparent_index = 0, delay = 0, image_disposal = 0;

  r = p;
  info->npics = 0;
  p->next = NULL;

  if ((GifFile = DGifOpen(info->ar, ungif_input_func)) == NULL) {
#ifdef DEBUG
    PrintGifError();
#endif
    return PROC_NOT;
  }

  sheight = GifFile->SHeight;
  info->format = FORMAT_NAME;

  if ((ScreenBuffer = (GifRowType *)
       calloc(sheight, sizeof(GifRowType *))) == NULL) {
    if (DGifCloseFile(GifFile) == GIF_ERROR)
      PrintGifError();
    return PROC_ERROR;
  }    

  size = GifFile->SWidth * sizeof(GifPixelType);
  if ((ScreenBuffer[0] = (GifRowType)calloc(sheight, size)) == NULL) {
    if (DGifCloseFile(GifFile) == GIF_ERROR)
      PrintGifError();
    return PROC_ERROR;
  }
  for (i = 1; i < sheight; i++)
    ScreenBuffer[i] = ScreenBuffer[0] + i * size;

  do {
#ifdef DEBUG
    fprintf(stderr, "GetRecordType ");
#endif
    if (DGifGetRecordType(GifFile, &RecordType) == GIF_ERROR) {
      PrintGifError();
#if 0
      DGifCloseFile(GifFile);
      free(ScreenBuffer[0]);
      free(ScreenBuffer);
      return PROC_ERROR;
#endif
    }
#ifdef DEBUG
    fprintf(stderr, "done.\n");
#endif
    switch (RecordType) {
    case IMAGE_DESC_RECORD_TYPE:
      if (DGifGetImageDesc(GifFile) == GIF_ERROR)
	goto error;
      p->top = row = GifFile->Image.Top;
      p->left = col = GifFile->Image.Left;
      p->width = GifFile->Image.Width;
      p->height = GifFile->Image.Height;

#ifdef DEBUG
      fprintf(stderr, "IMAGE_DESC_RECORD_TYPE: (%d, %d) (%d, %d)\n",
	      p->left, p->top, p->width, p->height);
#endif

      memset(ScreenBuffer[0], GifFile->SBackGroundColor, sheight * size);
      if (GifFile->Image.Interlace) {
	for (i = 0; i < 4; i++)
	  for (j = row + ioffset[i]; j < row + p->height; j += ijumps[i])
	    if (DGifGetLine(GifFile, &ScreenBuffer[j][col], p->width) == GIF_ERROR)
	      goto error;
      } else {
	for (i = 0; i < p->height; i++)
	  if (DGifGetLine(GifFile, &ScreenBuffer[row++][col], p->width) == GIF_ERROR)
	    goto error;
      }

      p->type = _INDEX;
      p->bytes_per_line = p->width;
      p->image_size = p->bytes_per_line * p->height;
      p->background.index = GifFile->SBackGroundColor;
      if (info->npics == 0 && image_disposal == _RESTOREBACKGROUND) {
	/* guess... */
	p->transparent_disposal = _TRANSPARENT;
      } else {
	p->transparent_disposal = transparent_disposal;
      }
      p->transparent.index = transparent_index;
      p->image_disposal = image_disposal;
      p->delay = delay ? delay : 1;

      if ((p->image = malloc(p->image_size)) == NULL)
	goto error_after_close;

      for (i = 0; i < p->height; i++)
	memcpy(&p->image[i * p->width], &ScreenBuffer[p->top + i][p->left], p->width * sizeof(GifPixelType));

      ColorMap = GifFile->Image.ColorMap ? GifFile->Image.ColorMap : GifFile->SColorMap;
      p->ncolors = ColorMap->ColorCount;

      for (i = 0; i < p->ncolors; i++) {
	p->colormap[i][0] = ColorMap->Colors[i].Red;
	p->colormap[i][1] = ColorMap->Colors[i].Green;
	p->colormap[i][2] = ColorMap->Colors[i].Blue;
      }

      info->npics++;
      q = p;
      p->next = image_allocate();
      p = p->next;

      break;
    case EXTENSION_RECORD_TYPE:
      if (DGifGetExtension(GifFile, &extcode, &Extension) == GIF_ERROR)
	goto error;

      switch (extcode) {
      case COMMENT_EXT_FUNC_CODE:
	info->comment = malloc(Extension[0] + 1);
	memcpy(info->comment, &Extension[1], Extension[0]);
	info->comment[Extension[0]] = '\0';
	break;
      case GRAPHICS_EXT_FUNC_CODE:
	if (Extension[1] & 1) {
	  transparent_disposal = (info->npics > 0) ? _TRANSPARENT : info->transparent_disposal;
	  transparent_index = Extension[4];
	} else {
	  transparent_disposal = _DONOTHING;
	  transparent_index = -1;
	}
	image_disposal = (Extension[1] & 0x1c) >> 2;
	delay = Extension[2] + (Extension[3] << 8);
	break;
      default:
	break;
      }

      while (Extension != NULL) {
	if (DGifGetExtensionNext(GifFile, &Extension) == GIF_ERROR)
	  goto error;
	switch (extcode) {
	case COMMENT_EXT_FUNC_CODE:
	  /* Append comment... */
	  break;
	default:
	  break;
	}
      }
      break;
    default:
      break;
    }
  } while (RecordType != TERMINATE_RECORD_TYPE);

  if (DGifCloseFile(GifFile) == GIF_ERROR) {
    PrintGifError();
    goto error_after_close;
  }

  free(ScreenBuffer[0]);
  free(ScreenBuffer);

  image_destroy(p);
  q->next = NULL;
  if (info->npics > 1) {
    info->if_animated = 1;
    if (r->transparent_disposal == _SHAPE)
      r->transparent_disposal = _TRANSPARENT;
  } else {
    info->if_animated = 0;
  }

  return PROC_OK;

 error:
  PrintGifError();
  DGifCloseFile(GifFile);
 error_after_close:
  free(ScreenBuffer[0]);
  free(ScreenBuffer);
  if (p->next != NULL)
    image_destroy(p->next);

  return PROC_ERROR;
}

/* Static variable for signal handler */
static Info *global_info;

static void
ungif_alarmset(Info *info)
{
  struct itimerval value;
  Image *p = info->shown_image;

  global_info = info;

  if (p->delay > 0) {
    long delay = p->delay * 10000 * info->delay_factor;
    long delaysec = delay / 1000000, delayusec = delay % 1000000;

    if (signal(SIGALRM, (void(*)())info->handler.alarm) == SIG_ERR) {
      perror("ungif_alarmset(): signal()");
      exit(1);
    }
    value.it_interval.tv_sec = value.it_interval.tv_usec = 0L;
    value.it_value.tv_sec = delaysec;
    value.it_value.tv_usec = delayusec;
    if (setitimer(ITIMER_REAL, &value, NULL) != 0) {
      fprintf(stderr, "p->delay = %d, delay_factor = %f\n", p->delay, info->delay_factor);
      perror("ungif_alarmset(): setitimer()");
      exit(1);
    }
  }
}

static void *
ungif_alarm(int sig)
{
  global_info->rerender =
    (global_info->shown_image->next == NULL) ?
    RERENDER_FIRST : RERENDER_NEXT;

  return NULL;
}

static void
ungif_alarmoff(void)
{
  struct itimerval value;

  if (global_info != NULL && global_info->shown_image->delay) {
    value.it_value.tv_sec = 0;
    value.it_value.tv_usec = 0;
    setitimer(ITIMER_REAL, &value, NULL);
    signal(SIGALRM, SIG_DFL);
  }
}

int
ungif_load_image(Info *info, Image *p)
{
  int c;

  /* decode */
  if ((c = ungif_decode_image(info, p)) != PROC_OK)
    return c;

  /* set handler */
  info->handler.destroy  = NULL;
  info->handler.alarmset = ungif_alarmset;
  info->handler.alarm    = ungif_alarm;
  info->handler.alarmoff = ungif_alarmoff;

  return PROC_OK;
}
